import os
import sys
import threading
import json
import zipfile
import requests
import urllib3.exceptions
from retrying import retry
from tqdm import tqdm
import arrow
from mdai.preprocess import Project


def retry_on_http_error(exception):
    valid_exceptions = [
        requests.exceptions.HTTPError,
        requests.exceptions.ConnectionError,
        urllib3.exceptions.HTTPError,
    ]
    return any([isinstance(exception, e) for e in valid_exceptions])


class Client:
    """Client for communicating with MD.ai backend API.
    Communication is via user access tokens (in MD.ai Hub, Settings -> User Access Tokens).
    """

    def __init__(self, domain="public.md.ai", access_token=None):
        if not domain.endswith((".md.ai")):
            raise ValueError("domain {} is invalid: should be format *.md.ai".format(domain))

        self.domain = domain
        self.access_token = access_token
        self.session = requests.Session()
        self._test_endpoint()

    def project(self, project_id, path="."):
        """Initializes Project class given project id.

        Arguments:
            project_id: hash ID of project.
            path: directory used for data.
        """
        if path == ".":
            print("Using working directory for data.")
        else:
            os.makedirs(path, exist_ok=True)
            print("Using path '{}' for data.".format(path))

        data_manager_kwargs = {
            "domain": self.domain,
            "project_id": project_id,
            "path": path,
            "session": self.session,
            "headers": self._create_headers(),
        }
        annotations_data_manager = ProjectDataManager("annotations", **data_manager_kwargs)
        images_data_manager = ProjectDataManager("images", **data_manager_kwargs)

        annotations_data_manager.create_data_export_job()
        images_data_manager.create_data_export_job()

        annotations_data_manager.wait_until_ready()
        images_data_manager.wait_until_ready()

        p = Project(
            annotations_fp=annotations_data_manager.data_path,
            images_dir=images_data_manager.data_path,
        )
        return p

    def _create_headers(self):
        headers = {}
        if self.access_token:
            headers["x-access-token"] = self.access_token
        return headers

    def _test_endpoint(self):
        """Checks endpoint for validity and authorization.
        """
        test_endpoint = "https://{}/api/test".format(self.domain)
        r = self.session.get(test_endpoint, headers=self._create_headers())
        if r.status_code == 200:
            print("Successfully authenticated to {}.".format(self.domain))
        else:
            print("Authorization error. Make sure your access token is valid.")

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _gql(self, query, variables=None):
        """Executes GraphQL query.
        """
        gql_endpoint = "https://{}/api/graphql".format(self.domain)
        headers = self._create_headers()
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"

        data = {"query": query, "variables": variables}
        r = self.session.post(gql_endpoint, headers=headers, json=data)
        if r.status_code != 200:
            r.raise_for_status()

        body = r.json()
        data = body["data"] if "data" in body else None
        errors = body["errors"] if "errors" in body else None

        return data, errors


class ProjectDataManager:
    """Manager for project data exports and downloads.
    """

    def __init__(self, type, domain=None, project_id=None, path=".", session=None, headers=None):
        if type not in ["images", "annotations"]:
            raise ValueError("type must be 'images' or 'annotations'.")
        if not domain:
            raise ValueError("domain is not specified.")
        if not project_id:
            raise ValueError("project_id is not specified.")
        if not os.path.exists(path):
            raise OSError("Path '{}' does not exist.".foramt(path))

        self.type = type
        self.domain = domain
        self.project_id = project_id
        self.path = path
        if session and isinstance(session, requests.Session):
            self.session = session
        else:
            self.session = requests.Session()
        self.headers = headers

        # path for downloaded data
        self.data_path = None
        # ready threading event
        self._ready = threading.Event()

    def create_data_export_job(self):
        """Create data export job through MD.ai API.
        This is an async operation. Status code of 202 indicates successful creation of job.
        """
        endpoint = "https://{}/api/data-export/{}".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code == 202:
            msg = "Preparing {} export for project {}...".format(self.type, self.project_id)
            print(msg.ljust(100))
            self._check_data_export_job_progress()
        else:
            self._on_data_export_job_error()

    def wait_until_ready(self):
        self._ready.wait()

    def _get_data_export_params(self):
        if self.type == "images":
            params = {"projectHashId": self.project_id, "exportFormat": "zip"}
        elif self.type == "annotations":
            # TODO: restrict to assigned labelgroup
            params = {
                "projectHashId": self.project_id,
                "labelGroupNum": None,
                "exportFormat": "json",
            }
        return params

    def _check_data_export_job_progress(self):
        """Poll for data export job progress.
        """
        endpoint = "https://{}/api/data-export/{}/progress".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)

        try:
            body = r.json()
            status = body["status"]
            progress = body["progress"]
            time_remaining = body["timeRemaining"]
        except (TypeError, KeyError):
            self._on_data_export_job_error()
            return

        if body["status"] == "running":
            progress = int(progress) if progress else 0
            time_remaining = int(time_remaining) if time_remaining else 0
            # print formatted progress info
            if progress > 0 and progress <= 100 and time_remaining > 0:
                if time_remaining > 45:
                    time_remaining_fmt = (
                        arrow.now().shift(seconds=time_remaining).humanize(only_distance=True)
                    )
                else:
                    # arrow humanizes <= 45 to 'in seconds' or 'just now',
                    # so we will opt to be explicit instead.
                    time_remaining_fmt = "in {} seconds".format(time_remaining)
                end_char = "\r" if progress < 100 else "\n"
                msg = "Exporting {} for project {}...{}% (time remaining: {}).".format(
                    self.type, self.project_id, progress, time_remaining_fmt
                )
                print(msg.ljust(100), end=end_char, flush=True)
            # run progress check at 1s intervals so long as status == 'running' and progress < 100
            if progress < 100:
                t = threading.Timer(1.0, self._check_data_export_job_progress)
                t.start()
            return
        elif body["status"] == "done":
            self._on_data_export_job_done()
        elif body["status"] == "error":
            self._on_data_export_job_error()

    def _on_data_export_job_done(self):
        endpoint = "https://{}/api/data-export/{}/done".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        try:
            file_key = r.json()["fileKey"]
            if file_key:
                # download in separate thread
                t = threading.Thread(target=self._download_file, args=(file_key,))
                t.start()
        except (TypeError, KeyError):
            self._on_data_export_job_error(type, project_id)

    def _on_data_export_job_error(self):
        endpoint = "https://{}/api/data-export/{}/error".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        print("Error exporting {} for project {}.".format(self.type, self.project_id))

    def _download_file(self, file_key):
        """Downloads file via signed URL requested from MD.ai API.
        """
        print("Downloading file: {}".format(file_key))
        filepath = os.path.join(self.path, file_key)

        url = "https://{}/api/project-files/signedurl/get?key={}".format(
            self.domain, requests.utils.quote(file_key)
        )

        # stream response so we can display progress bar
        r = requests.get(url, stream=True, headers=self.headers)

        # total size in bytes
        total_size = int(r.headers.get("content-length", 0))
        block_size = 32 * 1024
        wrote = 0
        with open(filepath, "wb") as f:
            with tqdm(total=total_size, unit="B", unit_scale=True, unit_divisor=1024) as pbar:
                for chunk in r.iter_content(block_size):
                    f.write(chunk)
                    wrote = wrote + len(chunk)
                    pbar.update(block_size)
        if total_size != 0 and wrote != total_size:
            raise IOError("Error downloading file {}.".format(file_key))

        if self.type == "images":
            # unzip archive
            print("Extracting archive: {}".format(file_key))
            with zipfile.ZipFile(filepath, "r") as f:
                f.extractall(self.path)
            # images directory should have same name as archive file
            self.data_path = os.path.splitext(filepath)[0]
        elif self.type == "annotations":
            self.data_path = filepath

        # fire ready threading.Event
        self._ready.set()
        print("Success: {} data for project {} ready.".format(self.type, self.project_id))



import random


def hex2rgb(h):
    """Convert Hex color encoding to RGB color"""
    h = h.lstrip("#")
    return tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))


def train_test_split(image_ids, shuffle=True, validation_split=0.1):
    """Split image ids into training and validation sets."""
    if validation_split < 0.0 or validation_split > 1.0:
        return None, None

    image_ids_list = list(image_ids)
    if shuffle:
        sorted(image_ids_list)
        random.seed(42)
        random.shuffle(image_ids_list)

    split_index = int((1 - validation_split) * len(image_ids_list))
    image_ids_train = image_ids_list[:split_index]
    image_ids_val = image_ids_list[split_index:]

    print(
        "Num of instances for training set: %d, validation set: %d"
        % (len(image_ids_train), len(image_ids_val))
    )
    return image_ids_train, image_ids_val



import numpy as np
import re
import os
import warnings
import json
import pandas as pd
import pydicom


class Project:
    """Project consists of label groups, and datasets.
    """

    def __init__(self, annotations_fp=None, images_dir=None):
        """ 
        Args: 
            annotations_fp (str): File path to the exported JSON annotation file. 
            images_dir (str): File path to the DICOM images directory. 
        """
        self.annotations_fp = None
        self.images_dir = None
        self.label_groups = []
        self.datasets = []

        if annotations_fp is not None and images_dir is not None:
            self.annotations_fp = annotations_fp
            self.images_dir = images_dir

            with open(self.annotations_fp, "r") as f:
                self.data = json.load(f)

            for dataset in self.data["datasets"]:
                self.datasets.append(Dataset(dataset, images_dir))

            for label_group in self.data["labelGroups"]:
                self.label_groups.append(LabelGroup(label_group))
        else:
            print("Error: Missing data or images file paths!")

    def get_label_groups(self):
        return self.label_groups

    def show_label_groups_info(self):
        print("Label Groups:")
        for label_group in self.label_groups:
            print("Name: %s, Id: %s" % (label_group.name, label_group.id))
            print("\t")
            label_group.show_labels_info()
        print("")

    def get_label_group_by_name(self, label_group_name):
        for label_group in self.label_groups:
            if label_group.name == label_group_name:
                return label_group
        return None

    def get_label_group_by_id(self, label_group_id):
        for label_group in self.label_groups:
            if label_group.id == label_group_id:
                return label_group
        return None

    def get_datasets(self):
        return self.datasets

    def show_datasets_info(self):
        print("Datasets:")
        for dataset in self.datasets:
            print("Name: %s, Id: %s" % (dataset.name, dataset.id))
        print("")

    def get_dataset_by_name(self, dataset_name):
        for dataset in self.datasets:
            if dataset.name == dataset_name:
                return dataset
        return None

    def get_dataset_by_id(self, dataset_id):
        for dataset in self.datasets:
            if dataset.id == dataset_id:
                return dataset
        return None

    def _check_label_ids(self, label_ids):
        pass

    def set_label_ids(self, selected_label_ids):
        self.selected_label_ids = selected_label_ids
        self.labels_dict = self._create_labels_dict()

        for dataset in self.datasets:
            dataset.selected_label_ids = selected_label_ids
            dataset.labels_dict = self.labels_dict

    def _create_labels_dict(self):
        """Create a dict with label id as key, and a nested dict of class_id, and class_text 
        as values, e.g., {'L_v8n': {'class_id': 1, 'class_text': 'Lung Opacity'}}, 
        where L_v8n is the label id, with a class_id of 1 and class text of 'Lung Opacity'.

        Args: 
            label_ids (list): list of label ids. 
            labels (list): list of tuples of (label_id, label_name)
        
        Returns: 
            Label ids dictionary.
        """
        label_ids = self.selected_label_ids

        label_ids_dict = {}
        for i, label_id in enumerate(label_ids):

            # find which label group has the label
            for label_group in self.label_groups:
                labels = label_group.get_labels()
                for label in labels:
                    if label[0] == label_id:
                        class_text = label[1]
            label_ids_dict[label_id] = {"class_id": i, "class_text": class_text}
        return label_ids_dict


class LabelGroup:
    """A label group contains multiple labels.  
    
    Each label has properties such id, name, color, type, scope, annotation mode, rad lex tag ids.
    
    Label type: Global typed annotations apply to the whole instance (e.g., a CT image),
    while local typed annotations apply to a part of the image (e.g., ROI bounding box).

    Label scope: Scope can be of study, series, or instance.

    Label annotation mode: can be of bounding boxes, free form , polgon, etc.
    """

    def __init__(self, label_group):
        """
        Args: 
            label_group (object: json) JSON data for label group
        """
        self.label_group_data = label_group
        self.name = self.label_group_data["name"]
        self.id = self.label_group_data["id"]

    def get_labels(self):
        """Get label ids and names """
        return [(label["id"], label["name"]) for label in self.label_group_data["labels"]]

    def show_labels_info(self):
        """Show labels info"""
        print("Labels:")
        for label in self.label_group_data["labels"]:
            print("Name: %s, Id: %s" % (label["name"], label["id"]))
        print("")

    def get_label_group_dict(self):
        """Raw label group JSON data"""
        return self.label_group_data


class Dataset:
    """
    A dataset consists of DICOM images and annotations.
    """

    def __init__(self, dataset_data, images_dir):
        """ 
        Args: 
            dataset_data: Dataset json data
            images_dir: DICOM images directory. 
        """
        self.dataset_data = dataset_data
        self.images_dir = images_dir

        self.id = dataset_data["id"]
        self.name = dataset_data["name"]
        self.annotations = dataset_data["annotations"]
        self.selected_label_ids = None
        self.labels_dict = None

    def prepare(self):
        if self.selected_label_ids is None:
            raise Exception(
                "Label ids must be selected. Use select_label_ids() to \
                             set project wide label ids."
            )

        ann = self.get_annotations(self.selected_label_ids)
        self.imgs_anns = self._associate_images_and_annotations(ann)

    def get_annotations(self, label_ids):
        """Returns annotations, filtered by label ids.
        """
        if label_ids is None:
            print("Dataset contains %d annotations." % len(self.annotations))
            return self.annotations

        # TODO: print label ids
        ann_filtered = [a for a in self.annotations if a["labelId"] in label_ids]
        print("Dataset contains %d annotations." % len(ann_filtered))
        return ann_filtered

    def _generate_uid(self, ann):
        """Generate an unique image identifier based on the file path and DICOM structure.

        Args:
            ann (list): List of annotations.
        
        Returns:
            A unique image id.
        """
        uid = None
        try:
            uid = (
                os.path.join(
                    self.images_dir,
                    ann["StudyInstanceUID"],
                    ann["SeriesInstanceUID"],
                    ann["SOPInstanceUID"],
                )
                + ".dcm"
            )
        except Exception as error:
            print("Exception:", error)
            print("ann %s" % ann)
        return uid

    def get_image_ids(self):
        return self.image_ids

    def _get_image_ids(self, ann):
        """Get images ids for annotations.

        Args: 
            ann (list): List of annotations.

        Returns: 
            A list of image ids.
        """
        image_ids = set()
        for a in ann:
            uid = self._generate_uid(a)
            if uid:
                image_ids.add(uid)
        return list(image_ids)

    def _associate_images_and_annotations(self, ann):
        """Build a dictionary with image ids and annotations, filtered by label ids.

        Args: 
            ann (list): List of annotations.
            label_ids (list): List of label ids. Annotations is filtered based on label ids.
        
        Returns:
            Dictionary with image ids as keys and annotations as values.
        """
        self.image_ids = self._get_image_ids(ann)

        # empty dictionary with image ids as keys
        imgs_anns_dict = {fp: [] for fp in self.image_ids}
        for a in ann:
            uid = self._generate_uid(a)
            imgs_anns_dict[uid].append(a)
        return imgs_anns_dict

    def class_id_to_class_text(self, class_id):
        for k, v in self.labels_dict.items():
            if v["class_id"] == class_id:
                return v["class_text"]

